--- arr3
-- @author Enno Sylvester
-- @copyright 2017, 2018, 2019, 2020, 2021, 2023
-- @module eatglobal.arr3

local historie = [[
	Version 1.7:
	 - Added: arr3.div, arr3.max, arr3.calcSteps, arr3.split
	 - Changed: struct versionsData hinzugefügt
	Version 1.8:
	 - Added: Added Fields "__debug__" and "version" to result-table arr3.
	Version 1.9:
	 - Added: Functions angleUnit, calcMax, calcMin, calcNSectors, calcSteps, calcZArgAndAngel, copy, cross, degToRad, distance, getMiddle,
						getZAngel, getZData, dot, length, lengthTo, max, normalize, projectZPoint, rotateX, rotateY, rotateZ, split, toParams, 
						toArray and xyAngle now are Class-Function.
	 - Changed: arr3.new now accepts arr2, arr3, vec2, vec3, vec4 and numbers as parameters.
	 - Changed: +, -, * and / now accept both arr3 and numbers.
	 - Changed: Functions with table returns aka lists, returns now as eatSetEx.
	 - Changed: Deletion of all import functions, because arr3.new covers all possibilities.
	 - Changed: Deletion of the functions arr3.copy and arr3.mt, because arr3.new covers these functionalities. Class function copy is retained.
	Version 1.10:
	 - Added: Modulfähigkeit (Erweiterungsmöglichkeiten durch Module) implementiert.
	 - Changed: Switch to new system of initialization eatglopal-API (postLoadFn).
]]

local verionsData = {
	majorVersion = 1,
	minorVersion = 10,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local modules

local vec3
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		vec3 = global.vec3
	end,
	------------------------------------------------------------------------------------
	--	registerModule
	--	
	------------------------------------------------------------------------------------
	registerModule = function(moduleData)
		for modul in moduleData:_pairs() do
			if (type(modul.handler) == "table") then
				if (not modules) then
					modules = eatSetEx.new()
				end
				if modules:insertUnique(modul, function(m1, m2) return (m1.metaData.name == m2.metaData.name) end) then
					if (type(modul.init) == "function") then modul.init() end
				end
			end
		end
	end,
}

------------------------------------------------------------------------------------
--	Vordefinition der "functionClass".
------------------------------------------------------------------------------------
local functionClass

------------------------------------------------------------------------------------
--	
------------------------------------------------------------------------------------
local function getArr3(x, y, z)
	if (type(x) == "table") then
		if ((type(x[1]) == "number") and (type(x[2]) == "number")) then
			-- x is a arr2 or arr3
			return {x[1], x[2], (type(x[3]) == "number") and x[3] or 0}
		elseif ((type(x.x) == "number") and (type(x.y) == "number")) then
			-- x is a vec2, vec3 or vec4
			return {x.x, x.y, (type(x.z) == "number") and x.z or 0}
		end
	elseif (type(x) == "number") then
		-- x is a number; We look to see if y and z are also numbers. Otherwise they get x as value
		return {x, (type(y) == "number") and y or x, (type(z) == "number") and z or x}
	end
	-- nothing; return default
	return {0, 0, 0}
end

local function angleArg(arg)
	return math.acos(math.eat.clamp(arg, -1, 1))
end

local function getSinCos(rot)
	return math.sin(rot), math.cos(rot)
end

local mt
mt = {
  __add = function (self, rhs) -- "add" event handler
		if (type(rhs) == "number") then
			return setmetatable({ self[1] + rhs, self[2] + rhs, self[3] + rhs }, mt)
		else
			local v = getArr3(rhs)
			return setmetatable({ self[1] + v[1], self[2] + v[2], self[3] + v[3] }, mt)
		end
  end,
	__sub = function (self, rhs) -- "sub" event handler
		if (type(rhs) == "number") then
			return setmetatable({ self[1] - rhs, self[2] - rhs, self[3] - rhs }, mt)
		else
			local v = getArr3(rhs)
			return setmetatable({ self[1] - v[1], self[2] - v[2], self[3] - v[3] }, mt)
		end
  end,
	__mul = function (self, rhs) -- "mul" event handler
		if (type(rhs) == "number") then
			return setmetatable({ self[1] * rhs, self[2] * rhs, self[3] * rhs }, mt)
		else
			local v = getArr3(rhs)
			return setmetatable({ self[1] * v[1], self[2] * v[2], self[3] * v[3] }, mt)
		end
	end,
	__div = function (self, rhs) -- "div" event handler
		if (type(rhs) == "number") then
			return setmetatable({ self[1] / rhs, self[2] / rhs, self[3] / rhs }, mt)
		else
			local v = getArr3(rhs)
			return setmetatable({ self[1] / v[1], self[2] / v[2], self[3] / v[3] }, mt)
		end
	end,
	__index = {
		angleUnit = function(self, p)
			return angelArg(self:dot(p))
		end,
		calcMax = function(self, p)
			return setmetatable({math.max(self[1], p[1]), math.max(self[2], p[2]), math.max(self[3], p[3])}, mt)
		end,
		calcMin = function(self, p)
			return setmetatable({math.min(self[1], p[1]), math.min(self[2], p[2]), math.min(self[3], p[3])}, mt)
		end,
		calcNSectors = function(self, p, countSectors)
			local result = eatSetEx.new({self:copy()})
			local pEnd = functionClass.new(p)
			local pSector = (pEnd - self) / countSectors
			for i = 1, countSectors - 1 do
				result:insert(self + (pSector * i))
			end
			result:insert(pEnd)
			
			return result
		end,
		calcSteps = function(self, p, range)
			return math.round(math.max((self - p):max(), range) / range)
		end,
		calcZArgAndAngel = function(self, p, radius)
			local arg, angel, length = self:getZData(p, radius)
			return arg, angel * length
		end,
		copy = function(self)
			return setmetatable({self:toParams()}, mt)
		end,
		cross = function(self, p)
			return setmetatable({
				self[2] * p[3] - self[3] * p[2],
				self[3] * p[1] - self[1] * p[3],
				self[1] * p[2] - self[2] * p[1]
			}, mt)
		end,
		degToRad = function(self)
			return setmetatable({math.rad(self[1]), math.rad(self[2]), math.rad(self[3])}, mt)
		end,
		distance = function(self, p)
			return (self - p):length()
		end,
		getMiddle = function(self, p)
			return self + ((functionClass.new(p) - self) / 2)
		end,
		getZAngel = function(self, p)
			local function calcTotalLength(yLength, z)
				local fac = (yLength < 0) and -1 or 1
				return math.sqrt(yLength^2 + z^2) * fac
			end
			
			local sx, sy, sz = self:toParams()
			local ex, ey, ez = functionClass.new(p):toParams()
			local z = math.abs(ez - sz)
			return z / calcTotalLength(ey - sy, z)
		end,
		getZData = function(self, p, radius)
			local arg = p[3] - self[3]
			return arg, angleArg(arg), self:lengthTo(p, radius)
		end,
		dot = function(self, p)
			return self[1] * p[1] + self[2] * p[2] + self[3] * p[3]
		end,
		length = function(self)
			return math.sqrt(self[1]^2 + self[2]^2 + self[3]^2)
		end,
		lengthTo = function(self, p, radius)
			local r = radius or 0
			local l = math.abs(self:distance(p))
			if (r == 0) then
				return l
			else
				return 2 * r * math.asin( l / (2 * r) )
			end
		end,
		max = function(self)
			local result = 0
			for i, value in ipairs(self) do
				result = math.max(result, math.abs(value))
			end
			return result
		end,
		normalize = function(self)
			return self * (1 / self:length())
		end,
		projectZPoint = function(self, angel, yLength)
			return setmetatable({self[1], self[2] + yLength, self[3] + (math.tan(angel) * yLength)}, mt)
		end,
		--- alle rotate...-Funktionen mit rot in RAD
		-- Umrechnung: aRAD = math.rad(aDEG)
		rotateX = function(self, rot)
			local s, c = getSinCos(rot)
			return setmetatable({self[1], (self[2] * c - self[3] * s), (self[2] * s + self[3] * c)}, mt)
		end,
		rotateY= function(self, rot)
			local s, c = getSinCos(rot)
			return setmetatable({(self[1] * c + self[3] * s), self[2], (self[1] * -s + self[3] * c)}, mt)
		end,
		rotateZ= function(self, rot)
			local s, c = getSinCos(rot)
			return setmetatable({(self[1] * c - self[2] * s), (self[2] * s + self[2] * c), self[3]}, mt)
		end,
		split = function(self, p, steps, insertLast)
			local result = eatSetEx.new({self:copy()})
			local pEnd = functionClass.new(p)
			local addP = (pEnd - self) / steps
			local currentPoint = self:copy()
			for i = 1, steps - 1 do
				local nextPoint = currentPoint + addP
				result:insert(nextPoint)
				currentPoint = nextPoint
			end
			
			if ((type(insertLast) ~= "boolean") and true or insertLast) then
				result:insert(pEnd)
			end
			
			return result
		end,
		swap = function(self)
			return setmetatable({self[3], self[2], self[1]}, mt)
		end,
		toParams = function(self)
			return self[1], self[2], self[3]
		end,
		toVec = function(self)
			return vec3.new(self)
		end,
		toString = function(self)
			return tostring(self)
		end,
		xyAngle = function(self)
			return math.atan2(self[2], self[1])
		end,
	},
	__tostring = function(self)
		return string.format("{%.4f, %.4f, %.4f}", self:toParams())
	end,
}

------------------------------------------------------------------------------------
--	functionClass.
------------------------------------------------------------------------------------
functionClass = {
	------------------------------------------------------------------------------------
	--	new
	--	
	------------------------------------------------------------------------------------
	new = function(x, y, z)
		return setmetatable(getArr3(x, y, z), mt)
	end,
	------------------------------------------------------------------------------------
	--	mtPoints
	--	
	------------------------------------------------------------------------------------
	mtPoints = function(points)
		local result = eatSetEx.new()
		for i = 1, #points do
			result:insert(functionClass.new(points[i]))
		end
		return result
	end,
	------------------------------------------------------------------------------------
	--	add
	--	
	------------------------------------------------------------------------------------
	add = function(a, b)
		return functionClass.new(a) + b
	end,
	------------------------------------------------------------------------------------
	--	sub
	--	
	------------------------------------------------------------------------------------
	sub = function(a, b)
		return functionClass.new(a) - b
	end,
	------------------------------------------------------------------------------------
	--	mul
	--	
	------------------------------------------------------------------------------------
	mul = function(f, p)
		return functionClass.new(p) * f
	end,
	------------------------------------------------------------------------------------
	--	div
	--	
	------------------------------------------------------------------------------------
	div = function(f, p)
		return functionClass.new(p) / f
	end,
	------------------------------------------------------------------------------------
	--	dot
	--	
	------------------------------------------------------------------------------------
	dot = function(a, b)
		return functionClass.new(a):dot(b)
	end,
	------------------------------------------------------------------------------------
	--	cross
	--	
	------------------------------------------------------------------------------------
	cross = function(a, b)
		return functionClass.new(a):cross(b)
	end,
	------------------------------------------------------------------------------------
	--	length
	--	
	------------------------------------------------------------------------------------
	length = function(p)
		return functionClass.new(p):length()
	end,
	------------------------------------------------------------------------------------
	--	distance
	--	
	------------------------------------------------------------------------------------
	distance = function(a, b)
		return functionClass.new(a):distance(b)
	end,
	------------------------------------------------------------------------------------
	--	max
	--	
	------------------------------------------------------------------------------------
	max = function(p)
		return functionClass.new(p):max()
	end,
	------------------------------------------------------------------------------------
	--	lengthBetween
	--	
	------------------------------------------------------------------------------------
	lengthBetween = function(p0, p1, radius)
		return functionClass.new(p0):lengthTo(p1, radius)
	end,
	------------------------------------------------------------------------------------
	--	getMiddle
	--	
	------------------------------------------------------------------------------------
	getMiddle = function(p0, p1)
		return functionClass.new(p0):getMiddle(p1)
	end,
	------------------------------------------------------------------------------------
	--	calcSteps
	--	
	------------------------------------------------------------------------------------
	calcSteps = function(p0, p1, range)
		return functionClass.new(p0):calcSteps(p1, range)
	end,
	------------------------------------------------------------------------------------
	--	split
	--	
	------------------------------------------------------------------------------------
	split = function(p0, p1, steps, insertLast)
		return functionClass.new(p0):split(p1, steps, insertLast)
	end,
	------------------------------------------------------------------------------------
	--	calcNSectors
	--	
	------------------------------------------------------------------------------------
	calcNSectors = function(p0, p1, countSectors)
		return functionClass.new(p0):calcNSectors(p1, countSectors)
	end,
	------------------------------------------------------------------------------------
	--	getZAngel
	--	
	------------------------------------------------------------------------------------
	getZAngel = function(pStart, pEnd)
		return functionClass.new(pStart):getZAngel(pEnd)
	end,
	------------------------------------------------------------------------------------
	--	projectZPoint
	--	
	------------------------------------------------------------------------------------
	projectZPoint = function(p, angel, yLength)
		return functionClass.new(p):projectZPoint(angel, yLength)
	end,
	------------------------------------------------------------------------------------
	--	normalize
	--	
	------------------------------------------------------------------------------------
	normalize = function(p)
		return functionClass.new(p):normalize()
	end,
	------------------------------------------------------------------------------------
	--	normize
	--	
	------------------------------------------------------------------------------------
	normize = function(points)
		local result = eatSetEx.new()
		if (#points > 0) then
			local min = functionClass.new(points[1])
			local max = functionClass.new(points[1])
			local container = eatSetContainer(points)
			for point in container:_pairs() do
				min = min:calcMin(point)
				max = max:calcMax(point)
			end
			for point in container:_pairs() do
				local temp = functionClass.new()
				for i = 1, 3 do
					if (max[i] == min[i]) then
						temp[i] = 0
					else
						temp[i] = (point[i] - min[i]) / (max[i] - min[i])
					end
				end
				result:insert(temp)
			end
		end
		
		return result
	end,
	------------------------------------------------------------------------------------
	--	angleUnit
	--	
	------------------------------------------------------------------------------------
	angleUnit = function(a, b)
		return functionClass.new(a):angleUnit(b)
	end,
	------------------------------------------------------------------------------------
	--	xyAngle
	--	
	------------------------------------------------------------------------------------
	xyAngle = function(p)
		return functionClass.new(p):xyAngle()
	end,
	------------------------------------------------------------------------------------
	--	alle rotate...-Funktionen mit rot in RAD
	--	Umrechnung: aRAD = math.rad(aDEG)
	------------------------------------------------------------------------------------
	--	rotateX
	--	
	------------------------------------------------------------------------------------
	rotateX = function(p, rot)
		return functionClass.new(p):rotateX(rot)
	end,
	------------------------------------------------------------------------------------
	--	rotateY
	--	
	------------------------------------------------------------------------------------
	rotateY = function(p, rot)
		return functionClass.new(p):rotateY(rot)
	end,
	------------------------------------------------------------------------------------
	--	rotateZ
	--	
	------------------------------------------------------------------------------------
	rotateZ = function(p, rot)
		return functionClass.new(p):rotateZ(rot)
	end,
	------------------------------------------------------------------------------------
	--	degToRad
	--	
	------------------------------------------------------------------------------------
	degToRad = function(p)
		return functionClass.new(p):degToRad()
	end,
	------------------------------------------------------------------------------------
	--	getZData
	--	
	------------------------------------------------------------------------------------
	getZData = function(p0, p1, radius)
		return functionClass.new(p0):getZData(p1, radius)
	end,
	------------------------------------------------------------------------------------
	--	calcZArgAndAngel
	--	
	------------------------------------------------------------------------------------
	calcZArgAndAngel = function(p0, p1, radius)
		return functionClass.new(p0):calcZArgAndAngel(p1, radius)
	end,
	------------------------------------------------------------------------------------
	--	toVec
	--	
	------------------------------------------------------------------------------------
	-- Exportfunktionen
	toVec = function(p)
		return vec3.new(p)
	end,
	------------------------------------------------------------------------------------
	--	toParams
	--	
	------------------------------------------------------------------------------------
	toParams = function(pt)
		return pt[1], pt[2], pt[3]
	end,
}

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
local modulDB = setmetatable({}, {
	__index = function(self, key)
		if modules then
			for modul in modules:_pairs() do
				local value = modul.handler[key]
				if (value ~= nil) then
					return value
				end
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

return setmetatable({}, {
	__index = function(self, key)
		for i, class in ipairs({libResult, functionClass, modulDB}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})

