--- eatglobal_intern
-- @author Enno Sylvester
-- @copyright 2017, 2018, 2019, 2020, 2021
-- @module eatglobal.eatglobal_intern

local historie = [[
	Version 1.8:
	 - Added: Added to lib math: clamp, sign, clearNumberArray, splitNumberRange
	 - Added: Global functions _pcall, _xpcall, _pRequire
	 - Changed: Abfrage, ob Funktionen nicht bereits in lib math verfügbar.
	 - Changed: struct versionsData hinzugefügt
	Version 1.9:
	 - Changed: eatset ist ab jetzt eatSetEx
	 - Added: eatSetEx:insert(value, count)
	 - Added: eatSetEx:append(values)
	 - Added: eatSetEx:count(filterFn)
	 - Added: Globale Funktion eatCopyTable(tbl)
	 - Added: Globale Funktion eatCreateRep(_api)
	 - Added: Globale Funktion setVarByVersion(currentValue, settingValue, currentVersionsData, settingVersionsData)
	 - Changed: eatSetEx.new - Wenn Parameter keine Tabelle wird eine leere Tabelle zurück gegeben
	 - Changed: eatSetEx.new - Neuer Parameter "returnAsCopy". Nur bei "true" wird eine Kopie von "var" zurück gegeben.
	 - Changed: eatSetEx:map - Parameter von "fnc, filterFn" geändert zu "fnc, params, orderFn, filterFn" für benutzerdefinierte Parameter und sortierte Ausgaben.
	 - Changed: eatSetEx:forEach - Parameter von "fnc, filterFn" geändert zu "fnc, params, orderFn, filterFn" für benutzerdefinierte Parameter und sortierte Ausgaben.
	 - Changed: eatSetEx:_pairs - Füllen Tabelle "keys" nun per Index und nicht mehr über self.pairs.
	Version 1.10:
	 - Added: class eatContainer
	 - Added: function eatConcatTable(tbl, seperator)
	 - Added: function eatSortPairs(tbl, orderFn)
	 - Added: function string.eat.split(s, delimiter, plain)
	 - Added: function eatToString(object, prefix)
	 - Added: eatSetEx:asString(index)
	 - Changed: Korrektur in eatSetEx:tostring() --> Rückgabe von Strings nun in Anführungszeichen ("")
	 - Changed: Delete global function compareVersion (neue globale Funktion  eatCompareVersion nutzen!!).
	 - Changed: Delete global function setVarByVersion
	 - Changed: Alle math-Funktionen in separate Library eatglobal.math verschoben.
	Version 1.11:
	 - Added: function eatContainer:firstValue(value, compareFn, filterFn)
	 - Added: function eatSetEx:firstValue(value, compareFn, filterFn)
]]

------------------------------------------------------------------------------------
--	Laden benötigter Module, damit weitere require in den Modulen nicht nötig sind.
------------------------------------------------------------------------------------
require "stringutil"

------------------------------------------------------------------------------------
--	Aktuelle Versionsdaten.
------------------------------------------------------------------------------------
local versionsData = {
	majorVersion = 1,
	minorVersion = 11,
}

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult = {
	__debug__ = false,
	historie = historie,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback

local math
local table

------------------------------------------------------------------------------------
--	Erzeugt von jedem beliebigen Objekt einen String z.B. für Logausgaben.
------------------------------------------------------------------------------------
local function _eatToString(object, prefix)
	local stack = {}
	local function writeFn(...)
		stack[#stack + 1] = table.concat({...})
	end
	
	local function checkOneLine(tbl)
		local setKeys = {}
		local tableKeys = {}
		local oneLine = true
		for i in ipairs(tbl) do
			setKeys[i] = true
		end
		for key, value in pairs(tbl) do
			if (type(value) == "table") then oneLine = false end
			if not setKeys[key] then
				table.insert(tableKeys, key)
				oneLine = false
			end
		end
		table.sort(tableKeys)
		return oneLine, tableKeys
	end
	---------------------
	local function _do(obj, prefix)
		local function writeKey(key)
			if ((type(key) == "string") and string.find(key, "^[_%a][_%w]*$")) then
				writeFn(key, " = ")
			else
				writeFn("[")
				_do(key, "")
				writeFn("] = ")
			end
		end
		-------------------------
		local tv = type(obj)
		if (tv == "nil") then
			writeFn("nil")
		elseif (tv == "boolean") then
			writeFn(tostring(obj))
		elseif (tv == "number") then
			writeFn(obj)
		elseif (tv == "string") then
			writeFn(string.format("%q", obj))
		elseif (tv == "table") then
			local oneLine, tableKeys = checkOneLine(obj)
			if oneLine then
				writeFn("{ ")
				for i, value in ipairs(obj) do
					if (i > 1) then
						writeFn(", ")
					end
					_do(value, "")
					--writeFn(", ")
				end
				for i, key in ipairs(tableKeys) do
					local value = obj[key]
					writeKey(key)
					if (i > 1) then
						writeFn(", ")
					end
					_do(value, "")
					--writeFn(", ")	
				end
				writeFn(" }")
			else
				local prefix2 = prefix .. "\t"
				writeFn("{\n")
				for key, value in ipairs(obj) do
					writeFn(prefix2)
					_do(value, prefix2)
					writeFn(",\n")
				end
				for i, key in ipairs(tableKeys) do
					local value = obj[key]
					writeFn(prefix2)
					writeKey(key)
					_do(value, prefix2)
					writeFn(",\n")	
				end
				writeFn(prefix, "}")
			end
		elseif (tv == "function") then
			writeFn("function(...)")
		else
			writeFn('"', tv, '"')
		end
	end
	
	_do(object, ((type(prefix) == "string") and prefix or ""))
	return table.concat(stack)
end

------------------------------------------------------------------------------------
--	Class eatSetEx
------------------------------------------------------------------------------------
local set = {}
local mt_set = {
  __add = function(self, var)
    local result = set.new{}
    for i, v in ipairs(self) do
      result:insert((var[i] ~= nil) and _pcall(function() return v + var[i] end) or nil)
    end
    return result
  end,
  __sub = function(self, var)
    local result = set.new{}
    for i, v in ipairs(self) do
      result:insert((var[i] ~= nil) and _pcall(function() return v - var[i] end) or nil)
    end
    return result
  end,
  __mul = function(self, var)
    local result = set.new{}
    for i, v in ipairs(self) do
      result:insert((var[i] ~= nil) and _pcall(function() return v * var[i] end) or nil)
    end
    return result
  end,
  __div = function(self, var)
    local result = set.new{}
    for i, v in ipairs(self) do
      result:insert((var[i] ~= nil) and _pcall(function() return v / var[i] end) or nil)
    end
    return result
  end,
	__eq = function(self, var)
		if (#self == #var) then
			for i, v in ipairs(self) do
				if (v ~= var[i]) then
					return false
				end
			end
			return true
		else
			return false
		end
	end,
	__tostring = function(self)
		return eatToString(self)
	end,
	__index = {
		_pairs = function(self, orderFn, filterFn)
			-- collect the keys
			local keys = {}
			for i = 1, #self do
				keys[#keys + 1] = i
			end
			if orderFn then
				table.sort(keys, function(a, b) return orderFn(self[a], self[b]) end)
			end
			
			local doFilter = global.misc.buildFilterFn(filterFn)
			-- return the iterator function
			local i = 0
			return function()
				i = i + 1
				while keys[i] and (not doFilter(self[keys[i]])) do
					i = i + 1
				end
				if keys[i] then
					return self[keys[i]]
				end
			end
		end,
		ipairs = function(self, orderFn, filterFn)
			-- collect the keys
			local keys = {}
			for i = 1, #self do
				keys[#keys + 1] = i
			end
			if orderFn then
				table.sort(keys, function(a, b) return orderFn(self[a], self[b]) end)
			end
			
			local doFilter = global.misc.buildFilterFn(filterFn)
			-- return the iterator function
			local i = 0
			return function()
				i = i + 1
				while keys[i] and (not doFilter(self[keys[i]])) do
					i = i + 1
				end
				if keys[i] then
					return i, self[keys[i]]
				end
			end
		end,
		pairs = function(self, orderFn, filterFn)
			-- collect the keys
			local keys = {}
			for k in pairs(self) do
				keys[#keys + 1] = k
			end
			if orderFn then
				table.sort(keys, function(a, b) return orderFn(self[a], self[b]) end)
			end
			
			local doFilter = global.misc.buildFilterFn(filterFn)
			-- return the iterator function
			local i = 0
			return function()
				i = i + 1
				while keys[i] and (not doFilter(self[keys[i]])) do
					i = i + 1
				end
				if keys[i] then
					return keys[i], self[keys[i]]
				end
			end
		end,
		append = function(self, values)
			for i, val in ipairs(values) do
				self:insert(val)
			end
		end,
		asString = function(self, index)
			return eatToString(self[index])
		end,
		copy = function(self)
			return self:filter()
		end,
    contains = function(self, value, compareFn, filterFn)
			return (self:firstValue(value, compareFn, filterFn) ~= nil) and true or false
    end,
		count = function(self, filterFn)
			local result = 0
			local doFilter = global.misc.buildFilterFn(filterFn)
			for i = 1, #self do
				if doFilter(self[i]) then
					result = result + 1
				end
			end
			return result
		end,
		filter = function(self, filterFn)
      local result = set.new{}
			for value in self:_pairs(nil, filterFn) do
				result:insert(value)
			end
      return result
    end,
		------------------------------------------------------------------------------------
		--	firstValue
		--	Liefert ersten gefundenen
		------------------------------------------------------------------------------------
    firstValue = function(self, value, compareFn, filterFn)
			for val in self:_pairs(nil, filterFn) do
        if (type(compareFn) == "function") then
          if compareFn(val, value) then
            return val
          end
        else
          if (val == value) then
            return val
          end
        end
			end
      return nil
		end,
		forEach = function(self, fnc, params, orderFn, filterFn)
			for value in self:_pairs(orderFn, filterFn) do
				fnc(value, params)
			end
		end,
		getReversed = function(self, filterFn)
      local result = set.new{}
			local doFilter = global.misc.buildFilterFn(filterFn)
			for i = #self, 1, -1 do
				local value = self[i]
				if doFilter(value) then
					result:insert(value)
				end
			end
			return result
		end,
		insert = function(self, value, count)
			count = (type(count) == "number") and math.abs(math.round(count)) or 1
			for i = 1, count do
				self[#self + 1] = value
			end
		end,
		insertSorted = function(self, value, compareFn)
			self:insert(value)
			if (type(compareFn) == "function") then
				table.sort(self, compareFn)
			end
		end,
		insertUnique = function(self, value, compareFn, filterFn)
			if (not self:contains(value, compareFn, filterFn)) then
				self:insert(value)
				return true
			end
			return false
		end,
		intersection = function(self, var)
			local result = set.new{}
			for i, v in ipairs(self) do
				for j, w in ipairs(var) do
					if (v == w) then
						result:insert(v)
						break
					end
				end
			end
			return result
		end,
		map = function(self, fnc, params, orderFn, filterFn)
      local result = set.new{}
			for value in self:_pairs(orderFn, filterFn) do
				result:insert(fnc(value, params))
			end
      return result
    end,
		print = function(self)
			print(tostring(self))
		end,
		getSorted = function(self, orderFn, filterFn)
      local result = set.new{}
			for value in self:_pairs(orderFn, filterFn) do
				result:insert(value)
			end
      return result
    end,
		tostring = function (self)
			return tostring(self)
		end,
		union = function(self, var)
			local result = set.new{}
			for i, v in ipairs(self) do result:insert(v) end
			if (type(var) == "table") then
				for i, v in ipairs(var) do result:insert(v) end
			end
			return result
		end,
		unionTables = function(self, tables)
			local result = set.new{}
			for i, v in ipairs(self) do result:insert(v) end
			if (type(tables) == "table") then
				for i, _table in ipairs(tables) do
					if (type(_table) == "table") then
						for j, v in ipairs(_table) do result:insert(v) end
					end
				end
			end
			return result
		end,
	},
}
setmetatable(set, mt_set)

function set.new(var, returnAsCopy)
	var = (type(var) == "table") and var or {}
	if (returnAsCopy == true) then
		local result = {}
		setmetatable(result, mt_set)
		for i, v in ipairs(var) do result[i] = v end
		return result
	else
		setmetatable(var, mt_set)
		return var
	end
end

function set.add(a, b)
	return set.new(a) + b
end

function set.sub(a, b)
	return set.new(a) - b
end

function set.mul(a, b)
	return set.new(a) * b
end

function set.div(a, b)
	return set.new(a) / b
end

--

function set._pairs(a, orderFn, filterFn)
  return set.new(a):_pairs(orderFn, filterFn)
end

function set.contains(a, value, compareFn, filterFn)
	return set.new(a):contains(value, compareFn, filterFn)
end

function set.filter(a, filterFn)
  return set.new(a):filter(filterFn)
end

function set.forEach(a, fnc, filterFn)
  set.new(a):forEach(fnc, filterFn)
end

function set.map(a, fnc, filterFn)
  return set.new(a):map(fnc, filterFn)
end

function set.intersection(a, b)
	return set.new(a):intersection(b)
end

function set.tostring(a)
	return set.new(a):tostring()
end

function set.print(a)
  set.new(a):print()
end

function set.union(a, b)
	return set.new(a):union(b)
end

------------------------------------------------------------------------------------
--	class eatContainer
--	
------------------------------------------------------------------------------------
local function _eatSetContainer(content)
	local mt_Functions = {
		__index = {
			_pairs = function(self, orderFn, filterFn)
				-- collect the keys
				local keys = {}
				for i = 1, #content do
					keys[#keys + 1] = i
				end
				if orderFn then
					table.sort(keys, function(a, b) return orderFn(content[a], content[b]) end)
				end
				
				local doFilter = global.misc.buildFilterFn(filterFn)
				-- return the iterator function
				local i = 0
				return function()
					i = i + 1
					while keys[i] and (not doFilter(content[keys[i]])) do
						i = i + 1
					end
					if keys[i] then
						return content[keys[i]]
					end
				end
			end,
			ipairs = function(self, orderFn, filterFn)
				-- collect the keys
				local keys = {}
				for i = 1, #content do
					keys[#keys + 1] = i
				end
				if orderFn then
					table.sort(keys, function(a, b) return orderFn(content[a], content[b]) end)
				end
				
				local doFilter = global.misc.buildFilterFn(filterFn)
				-- return the iterator function
				local i = 0
				return function()
					i = i + 1
					while keys[i] and (not doFilter(content[keys[i]])) do
						i = i + 1
					end
					if keys[i] then
						return i, content[keys[i]]
					end
				end
			end,
			pairs = function(self, orderFn, filterFn)
				-- collect the keys
				local keys = {}
				for k in pairs(content) do
					keys[#keys + 1] = k
				end
				if orderFn then
					table.sort(keys, function(a, b) return orderFn(content[a], content[b]) end)
				end
				
				local doFilter = global.misc.buildFilterFn(filterFn)
				-- return the iterator function
				local i = 0
				return function()
					i = i + 1
					while keys[i] and (not doFilter(content[keys[i]])) do
						i = i + 1
					end
					if keys[i] then
						return keys[i], content[keys[i]]
					end
				end
			end,
			append = function(self, values)
				for i, val in ipairs(values) do
					self:insert(val)
				end
			end,
			asString = function(self, index)
				return eatToString(content[index])
			end,
			copy = function(self)
				return self:filter()
			end,
			contains = function(self, value, compareFn, filterFn)
				return (self:firstValue(value, compareFn, filterFn) ~= nil) and true or false
			end,
			count = function(self, filterFn)
				local result = 0
				local doFilter = global.misc.buildFilterFn(filterFn)
				for i = 1, #content do
					if doFilter(content[i]) then
						result = result + 1
					end
				end
				return result
			end,
			filter = function(self, filterFn)
				local result = _eatSetContainer({})
				for value in self:_pairs(nil, filterFn) do
					result:insert(value)
				end
				return result
			end,
			firstValue = function(self, value, compareFn, filterFn)
				for val in self:_pairs(nil, filterFn) do
					if (type(compareFn) == "function") then
						if compareFn(val, value) then
							return val
						end
					else
						if (val == value) then
							return val
						end
					end
				end
				return nil
			end,
			forEach = function(self, fnc, params, orderFn, filterFn)
				for value in self:_pairs(orderFn, filterFn) do
					fnc(value, params)
				end
			end,
			getContent = function(self)
				return content
			end,
			getReversed = function(self, filterFn)
				local result = _eatSetContainer({})
				local doFilter = global.misc.buildFilterFn(filterFn)
				for i = #content, 1, -1 do
					local value = content[i]
					if doFilter(value) then
						result:insert(value)
					end
				end
				return result
			end,
			insert = function(self, value, count)
				count = (type(count) == "number") and math.abs(math.round(count)) or 1
				for i = 1, count do
					content[#content + 1] = value
				end
			end,
			insertSorted = function(self, value, compareFn)
				self:insert(value)
				if (type(compareFn) == "function") then
					table.sort(self.content, compareFn)
				end
			end,
			insertUnique = function(self, value, compareFn, filterFn)
				if (not self:contains(value, compareFn, filterFn)) then
					self:insert(value)
					return true
				end
				return false
			end,
			map = function(self, fnc, params, orderFn, filterFn)
				local result = _eatSetContainer({})
				for value in self:_pairs(orderFn, filterFn) do
					result:insert(fnc(value, params))
				end
				return result
			end,
			print = function(self)
				print(tostring(self))
			end,
			getSorted = function(self, orderFn, filterFn)
				local result = _eatSetContainer({})
				for value in self:_pairs(orderFn, filterFn) do
					result:insert(value)
				end
				return result
			end,
			tostring = function(self)
				return tostring(self)
			end,
			union = function(self, var)
				local result = _eatSetContainer({})
				for i, v in ipairs(content) do result:insert(v) end
				if (type(var) == "table") then
					for i, v in ipairs(var) do result:insert(v) end
				end
				return result
			end,
			unionTables = function(self, tables)
				local result = _eatSetContainer({})
				for i, v in ipairs(content) do result:insert(v) end
				if (type(tables) == "table") then
					for i, _table in ipairs(tables) do
						if (type(_table) == "table") then
							for j, v in ipairs(_table) do result:insert(v) end
						end
					end
				end
				return result
			end,
		},
	}
	local functionClass =  setmetatable({}, mt_Functions)
	
	local mt = {
		__eq = function(self, var)
			if (#content == #var) then
				for i, v in ipairs(content) do
					if (v ~= var[i]) then
						return false
					end
				end
				return true
			else
				return false
			end
		end,
		__len = function(self)
			return #content
		end,
		__newindex = function(self, key, value)
			content[key] = value
		end,
		__index = function(self, key)
			for _, class in ipairs({functionClass, content}) do
				local value = class[key]
				if (value ~= nil) then
					return value
				end
			end
		end,
		__ipairs = function(self)
			return function(self, i)
				i = i + 1
				local value = content[i]
				if (value ~= nil) then
					return i, value
				end
			end, self, 0
    end,
		__pairs = function(self)
      return function(self, key)
        local value
        key, value = next(content, key)
        if (value ~= nil) then
          return key, value
        end
      end, self, nil
    end,
		__tostring = function(self)
			return eatToString(content)
		end,
	}
	
	content = global.misc.getValue(content, {})
	return setmetatable({}, mt)
end

------------------------------------------------------------------------------------
--	UNDER CONSTRUCTION!
--	class _eatTableContainer
------------------------------------------------------------------------------------
local function _eatTableContainer(content)
	local mt_Functions = {
		__index = {
			pairs = function(self, orderFn, filterFn)
				-- collect the keys
				local keys = {}
				for k in pairs(content) do
					keys[#keys + 1] = k
				end
				if orderFn then
					table.sort(keys, function(a, b) return orderFn(a, b) end)
				end
				
				local doFilter = global.misc.buildFilterFn(filterFn)
				-- return the iterator function
				local i = 0
				return function()
					i = i + 1
					while keys[i] and (not doFilter(content[keys[i]])) do
						i = i + 1
					end
					if keys[i] then
						return keys[i], content[keys[i]]
					end
				end
			end,
		},
	}
	local functionClass =  setmetatable({}, mt_Functions)
	---------------------------------------------------
	local mt = {
		__len = function(self)
			return #content
		end,
		__newindex = function(self, key, value)
			content[key] = value
		end,
		__index = function(self, key)
			for _, class in ipairs({functionClass, content}) do
				local value = class[key]
				if (value ~= nil) then
					return value
				end
			end
		end,
		__ipairs = function(self)
			return function(self, i)
				i = i + 1
				local value = content[i]
				if (value ~= nil) then
					return i, value
				end
			end, self, 0
    end,
		__pairs = function(self)
      return function(self, key)
        local value
        key, value = next(content, key)
        if (value ~= nil) then
          return key, value
        end
      end, self, nil
    end,
		__tostring = function(self)
			return eatToString(content)
		end,
	}
	---------------------------------------------------
	content = (type(content) == "table") and content or {}
	return setmetatable({}, mt)
end

------------------------------------------------------------------------------------
--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
------------------------------------------------------------------------------------
function libResult.postLoadFn(params)
	-------------------------------------------------------------------
	--	Lokale Variable "global", etc. zuweisen.
	-------------------------------------------------------------------
	global = params.eatglobal
	getVar = params.getVar
	setVar = params.setVar
	logFn = params.logFn
	logFnTraceback = params.logFnTraceback
	eatVersion = getVar("eat.intern.eatVersion")
	libResult.version = tostring(eatVersion.new(versionsData))
	
	-------------------------------------------------------------------
	--	Global functions.
	-------------------------------------------------------------------
	_pcall = setVar(
		function (fnc, ...)
			local args = nil
			if (select('#', ...) > 0) then
				args = {...}
			end
			local state, result_or_error = pcall((args~= nil) and (function() return fnc(table.unpack(args)) end) or fnc)
			return state and result_or_error or nil, result_or_error
		end,
		"_pcall")

	_xpcall = setVar(
		function (fnc, catch, ...)
			local args = nil
			if (select('#', ...) > 0) then
				args = {...}
			end
			local state, result_or_error = xpcall((args~= nil) and (function() return fnc(table.unpack(args)) end) or fnc, catch)
			return state and result_or_error or nil, result_or_error
		end,
		"_xpcall")

	_pRequire = setVar(
		function (modulName)
			local state, result_or_error = pcall(require, modulName)
			return state and result_or_error or nil, result_or_error
		end,
		"_pRequire")

	------------------------------------------------------------------------------------
	--	ACHTUNG REIHENFOLGE!!!
	--	Erst die globalen Variablen setzen und dann Bibliotheken wie math, table,
	--	etc. laden; andernfalls Absturzgefahr.
	------------------------------------------------------------------------------------
	eatToString = setVar(_eatToString, "eatToString")
	eatSetEx = setVar(set, "eatSetEx")
	eatSetContainer = setVar(_eatSetContainer, "eatSetContainer")
	eatTableContainer = setVar(_eatTableContainer, "eatTableContainer")
	
	------------------------------------------------------------------------------------
	------------------------------------------------------------------------------------
	math = global.math
	table = global.table
end

------------------------------------------------------------------------------------
------------------------------------------------------------------------------------
return libResult
