--- msg
-- @author Enno Sylvester
-- @copyright 2023
-- @module eatglobal.msg

local historie = [[
	Version 1.1:
	 - Release
]]

local versionsData = {
	majorVersion = 1,
	minorVersion = 1,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback

local table
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		table = global.table
	end,
}

------------------------------------------------------------------------------------
--	Vordefinition der "functionClass".
------------------------------------------------------------------------------------
local functionClass

------------------------------------------------------------------------------------
--	mt_msg.
------------------------------------------------------------------------------------
local mt_msg = {
	__index = {
		create = function(self, msgId)
			----------------------------------------
			-- Tools
			----------------------------------------
			local function checkString(s)
				return ((type(s) == "string") and (s ~= "") and (not s:find(" ", 1, true)))
			end
			----------------------------------------
			local function getMessages(message)
				local result = eatSetEx.new()
				if (type(message) == "table") then
					for msg in eatSetContainer(message):_pairs() do
						if checkString(msg) then
							result:insertUnique(msg)
						end
					end
				elseif checkString(message) then
					result:insert(message)
				end
				return result
			end
			----------------------------------------
			local function checkInit()
				-- globale Daten
				if (type(eat1963_global_data) ~= "table") then
					eat1963_global_data = {}
				end
				-- msg
				if (type(eat1963_global_data.msg) ~= "table") then
					eat1963_global_data.msg = {
						handler = {},
					}
				end
				-- msgId
				if (type(eat1963_global_data.msg.handler[msgId]) ~= "table") then
					eat1963_global_data.msg.handler[msgId] = {
						stdHandler = eatSetEx.new(),
						msgHandler = {},
					}
				end
			end
			----------------------------------------
			local function getHandler()
				checkInit()
				return eat1963_global_data.msg.handler[msgId]
			end
			----------------------------------------
			local function getStdHandler()
				return getHandler().stdHandler
			end
			----------------------------------------
			local function getMsgHandler(message)
				local handler = getHandler()
				if (type(handler.msgHandler[message]) ~= "table") then
					handler.msgHandler[message] = eatSetEx.new()
				end
				return handler.msgHandler[message]
			end
			----------------------------------------
			local function getRunFn(message)
				local handler = getHandler()
				return handler.stdHandler:union(handler.msgHandler[message])
			end
			----------------------------------------
			local mt = {
				__index = {
					getMsgId = function(self)
						return msgId
					end,
					registerStdFn = function(self, runFn)
						if (type(runFn) == "function") then
							getStdHandler():insert(runFn)
						end
					end,
					registerMsgFn = function(self, message, runFn)
						if (type(runFn) == "function") then
							for msg in getMessages(message):_pairs() do
								getMsgHandler(msg):insert(runFn)
							end
						end
					end,
					sendMsg = function(self, message, ...)
						if checkString(message) then
							for runFn in getRunFn(message):_pairs() do
								runFn(message, ...)
							end
						end
					end,
					sendMsgEx = function(self, message, ...)
						if checkString(message) then
							for runFn in getRunFn(message):_pairs() do
								local result = runFn(message, ...)
								if (result ~= nil) then
									return result
								end
							end
						end
						return nil
					end,
					sendQuestion = function(self, message, ...)
						if checkString(message) then
							local result = eatSetEx.new()
							for runFn in getRunFn(message):_pairs() do
								local res = runFn(message, ...)
								if (res ~= nil) then
									return result:insert(res)
								end
							end
						end
						return result
					end,
				},
			}
			----------------------------------------
			--- HIER geht's los!!!!
			----------------------------------------
			msgId = checkString(msgId) and msgId or "eat1963_std_msg_id"
			return setmetatable({}, mt)
		end,
	},
}
		
------------------------------------------------------------------------------------
--	functionClass.
------------------------------------------------------------------------------------
functionClass = setmetatable({}, mt_msg)

local mt = {
	__index = function(self, key)
		for i, class in ipairs({libResult, functionClass}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
}

return setmetatable({}, mt)