--- Parameter Utils
-- @author Enno Sylvester
-- @copyright 2020
-- @module eatglobal.paramutils

local historie = [[
	Version 1.0:
	 - Release
	Version 1.1:
	 - Changed:	Eliminate minor bugs.
	Version 1.2:
	 - Changed:	runAllFun übergibt neben "data" nun auch den selektierten Index beginnend bei 0.
	Version 1.3:
	 - Changed:	Switch to new system of initialization eatglopal-API (postLoadFn).
	Version 1.4:
	 - Added:	Parameter "order" implementiert.
]]

local versionsData = {
	majorVersion = 1,
	minorVersion = 4,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback

local math
local valuesToStrings
local translateCaptions
local getValue

------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		math = global.math
		
		libResult.tools = {
			valuesToStrings = valuesToStrings,
			makeRange = math.makeRange,
			makeMiddledRange = math.makeMiddledRange,
			translateArray = translateCaptions,
		}
		getValue = global.misc.getValue
	end,
}

------------------------------------------------------------------------------------
--	valuesToStrings
------------------------------------------------------------------------------------
valuesToStrings = function(values)
	local result = eatSetEx.new()
	for val in eatSetContainer(values):_pairs() do
		if (type(val) == "string") then
			result:insert(val)
		else
			result:insert(tostring(val))
		end
	end
	
	return result
end

------------------------------------------------------------------------------------
--	translateCaptions
------------------------------------------------------------------------------------
translateCaptions = function(captions)
	local result = {}
	for val in eatSetContainer(captions):_pairs() do
		table.insert(result, _(val))
	end
	
	return result
end

------------------------------------------------------------------------------------
--	create
------------------------------------------------------------------------------------
function libResult.create()
	------------------------------------------------------------------------------------
	--	makeEntry
	--
	------------------------------------------------------------------------------------
	local function makeEntry(result, uiType, key, name, values, defaultIndex, tooltip, data, fun, order)
		local function checkName(name)
			if ((type(name) == "string") and (name:sub(1, 1) == "*")) then
				return key..name:sub(2)
			else
				return name
			end
		end
		
		result.params:insertSorted({
					uiType = uiType,
					key = key,
					name = _(checkName(name)),
					values = values,
					defaultIndex = defaultIndex or 0,
					tooltip = _(checkName(tooltip)),
					order = getValue(order, result.currentDefaultOrder),
				}, function(a, b) return (a.order < b.order) end
			)
		
		result.currentDefaultOrder = result.currentDefaultOrder + 1
		result.data[key] = {
			data = (data ~= nil) and ((type(data) == "table") and data or {data}) or {},
			fun = (type(fun) == "function") and fun or nil,
		}
	end
	
	------------------------------------------------------------------------------------
	------------------------------------------------------------------------------------
	local result
	result = {
		------------------------------------------------------------------------------------
		--	currentDefaultOrder
		--	Ist "order" nicht angegeben, so wird "currentDefaultOrder" verwendet.
		--	"currentDefaultOrder" wird je Zugriff um 1 erhöht.
		------------------------------------------------------------------------------------
		currentDefaultOrder = 999,
		params = eatSetEx.new(),
		data = {},
		createButton = function(params, data, fun)
			makeEntry(result, "BUTTON", params.key, params.name, translateCaptions(valuesToStrings(params.values)),
				params.defaultIndex, params.tooltip, data, fun, params.order)
		end,
		createCheckbox = function(params, data, fun)
			makeEntry(result, "CHECKBOX", params.key, params.name, params.values or translateCaptions({"No", "Yes"}),
			params.isChecked and 1 or 0, nil, data, fun, params.order)
		end,
		createCombobox = function(params, data, fun)
			makeEntry(result, "COMBOBOX", params.key, params.name, translateCaptions(valuesToStrings(params.values)),
				params.defaultIndex, params.tooltip, data, fun, params.order)
		end,
		createIconButton = function(params, data, fun)
			makeEntry(result, "ICON_BUTTON", params.key, params.name, params.icons,
				params.defaultIndex, params.tooltip, data, fun, params.order)
		end,
		createSlider = function(params, data, fun)
			makeEntry(result, "SLIDER", params.key, params.name, valuesToStrings(params.values),
				params.defaultIndex, params.tooltip, data, fun, params.order)
		end,
	}
	result.makeButton = function(key, name, values, defaultIndex, tooltip, data, fun)
		makeEntry(result, "BUTTON", key, name, translateCaptions(valuesToStrings(values)), defaultIndex, tooltip, data, fun)
	end
	result.makeCheckbox = function(key, name, values, isChecked, data, fun)
		makeEntry(result, "CHECKBOX", key, name, values or translateCaptions({"No", "Yes"}), isChecked and 1 or 0, nil, data, fun)
	end
	result.makeCombobox = function(key, name, values, defaultIndex, tooltip, data, fun)
		makeEntry(result, "COMBOBOX", key, name, translateCaptions(valuesToStrings(values)), defaultIndex, tooltip, data, fun)
	end
	result.makeIconButton = function(key, name, icons, defaultIndex, tooltip, data, fun)
		makeEntry(result, "ICON_BUTTON", key, name, icons, defaultIndex, tooltip, data, fun)
	end
	result.makeSlider = function(key, name, values, defaultIndex, tooltip, data, fun)
		makeEntry(result, "SLIDER", key, name, valuesToStrings(values), defaultIndex, tooltip, data, fun)
	end
	
	result.getIndex = function(params, key) return params[key] or 0 end
	result.getData = function(params, key)
		return result.data[key].data[(params[key] or 0) + 1]
	end
	result.getValue = function(params, key)
		return result.data[key].values[(params[key] or 0) + 1]
	end
	result.runAllFun = function(params)
		for key, val in pairs(result.data) do
			if (type(val.fun) == "function") then
				local ix = (params[key] or 0)
				val.fun(val.data[ix + 1], ix, key)
			end
		end
	end
	
	return result
end

return libResult

