--- capacityhandler
-- @author Enno Sylvester
-- @copyright 2023
-- @module eatglobal.module.capacityhandler

local historie = [[
	Version 1.1:
	 - Release
	Version 1.2:
	 - Fixed: Sicherheitsabfrage Ermittlung capacityFactor hinzugefügt.
]]

local versionsData = {
	majorVersion = 1,
	minorVersion = 1,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local getValue

local table
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		getValue = global.misc.getValue
		
		table = global.table
	end,
}

------------------------------------------------------------------------------------
--	meta-class
------------------------------------------------------------------------------------
local metaClass
metaClass = {
	-------------------------------------------------------------------
	--	Tabelle metaData
	--	
	-------------------------------------------------------------------
	metaData = {
		name = "capacityHandler",
		library = "modutils",
		priority = 10,
	},
	-------------------------------------------------------------------
	--	init
	--	
	-------------------------------------------------------------------
	init = function()
	end,
	-------------------------------------------------------------------
	--	Tabelle handler
	--	
	-------------------------------------------------------------------
	handler = {
		-------------------------------------------------------------------
		--	createUI
		--	
		-------------------------------------------------------------------
		createUI = function(params)
			local userSelection = getValue(params.userSelection, {})
			if (type(userSelection.handleCapacity) == "table") then
				local handleCapacity = userSelection.handleCapacity
				------------------------------------------------------------------------------------
				--	Wir müssen schauen, ob überhaupt Dateien zum behandeln vorhanden sind.
				------------------------------------------------------------------------------------
				local intern = params.modutilsIntern
				local capacityFiles = table.openKey(intern, "moduleData.capacityhandler.capacityFiles", true, true)
				
				local blackList = eatSetContainer(handleCapacity.blackList)
				for fileName in intern.availableModelFiles:union(getValue(handleCapacity.files, {})):_pairs() do
					capacityFiles:insertUnique(fileName, function(a, b) return (a == b) end)
				end
				
				if (#capacityFiles > 0) then
					local paramsEntry = handleCapacity.paramsEntry
					paramsEntry.values = eatSetEx.new()
					for value in eatSetContainer(handleCapacity.capacityFactors):_pairs() do
						paramsEntry.values:insert(string.format("%.2fx", value))
					end
					
					logFn("createUI. paramsEntry = ", paramsEntry)
					if (paramsEntry.uiType == "BUTTON") then
						params.myParams.createButton(paramsEntry)
						--params.myParams.makeButton(paramsEntry.key, paramsEntry.name, values, paramsEntry.defaultIndex, paramsEntry.tooltip)
					elseif (paramsEntry.uiType == "SLIDER") then
						params.myParams.createSlider(paramsEntry)
						--params.myParams.makeSlider(paramsEntry.key, paramsEntry.name, values, paramsEntry.defaultIndex, paramsEntry.tooltip)
					else
						params.myParams.createCombobox(paramsEntry)
						--params.myParams.makeCombobox(paramsEntry.key, paramsEntry.name, values, paramsEntry.defaultIndex, paramsEntry.tooltip)
					end
				end
			end
		end,
		-------------------------------------------------------------------
		--	run
		--	
		-------------------------------------------------------------------
		run = function(params)
			-------------------------------------------------------------------
			--	HIER geht's los!!!! :)
			-------------------------------------------------------------------
			local userSelection = getValue(params.userSelection, {})
			if (type(userSelection.handleCapacity) == "table") then
				-------------------------------------------------------------------
				--	Tools
				-------------------------------------------------------------------
				local function doCompare(val, fileName)
					return string.ends(fileName, val)
				end
				
				local blackList = eatSetContainer(params.userSelection.handleCapacity.blackList)
				local intern = params.modutilsIntern
				local capacityFiles = table.openKey(intern, "moduleData.capacityhandler.capacityFiles", true, true)
				
				global.mod.addModifier("loadModel", 
					function(fileName, data)
						if (capacityFiles:contains(fileName, doCompare) and data.metadata and data.metadata.transportVehicle) then
							local handlerData = userSelection.handleCapacity
							local capacityFactor = handlerData.capacityFactors[intern.userParams[handlerData.paramsEntry.key] or 1]
							
							if data.metadata.transportVehicle.compartmentsList then
								--	handling transportVehicle.compartmentsList
								for listEntry in eatSetContainer(data.metadata.transportVehicle.compartmentsList):_pairs(nil, function(value) return (type(value.loadConfigs) == 'table') end)  do
									for loadConfig in eatSetContainer(listEntry.loadConfigs):_pairs(nil, function(value) return (type(value.cargoEntries) == 'table') end) do
										for cargoEntrie in eatSetContainer(loadConfig.cargoEntries):_pairs() do
											cargoEntrie.capacity = (cargoEntrie.capacity or 0) * capacityFactor
										end
									end
								end
							elseif data.metadata.transportVehicle.capacities then
								--	handling transportVehicle.capacities
								for capacity in eatSetContainer(data.metadata.transportVehicle.capacities):_pairs() do
									capacity.capacity = (capacity.capacity or 0) * capacityFactor
								end
							elseif data.metadata.transportVehicle.compartments then
								--	handling transportVehicle.compartments
								for compartment in eatSetContainer(data.metadata.transportVehicle.compartments):_pairs() do
									for sub1 in eatSetContainer(compartment):_pairs() do
										for sub2 in eatSetContainer(sub1):_pairs() do
											sub2.capacity = (sub2.capacity or 0) * capacityFactor
										end
									end
								end
								--[[
								for i, v in ipairs(data.metadata.transportVehicle.compartments) do
									for j, w in ipairs(v) do
										for k, x in ipairs(w) do
											x.capacity = (x.capacity or 0) * capacityFactor
										end
									end
								end
								]]
							end
						end
						
						return data
					end,
					metaClass.metaData.priority
				)
			end
		end,
	},
}

local mt = {
	__index = function(self, key)
		for i, class in ipairs({libResult, metaClass}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
}

return setmetatable({}, mt)