--- fileshandler
-- @author Enno Sylvester
-- @copyright 2023
-- @module eatglobal.module.fileshandler

local historie = [[
	Version 1.1:
	 - Release
	Version 1.2:
	 - Changed:	modelFiles nun auch (wieder) als einzelner String möglich.
]]

local versionsData = {
	majorVersion = 1,
	minorVersion = 2,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback

local getValue
local isValidId
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		
		getValue = global.misc.getValue
		isValidId = global.misc.isValidId
end,
}

------------------------------------------------------------------------------------
--	Tool(s)
------------------------------------------------------------------------------------
local function getAsSet(value)
	if value then
		return (type(value) ~= "table") and eatSetContainer({value}) or eatSetContainer(value)
	end
end

------------------------------------------------------------------------------------
--	meta-class
------------------------------------------------------------------------------------
local metaClass = {
	-------------------------------------------------------------------
	--	Tabelle metaData
	--	
	-------------------------------------------------------------------
	metaData = {
		name = "fileshandler",
		library = "modutils",
	},
	-------------------------------------------------------------------
	--	init
	--	
	-------------------------------------------------------------------
	init = function()
	end,
	-------------------------------------------------------------------
	--	Tabelle handler
	--	
	-------------------------------------------------------------------
	handler = {
		-------------------------------------------------------------------
		--	createUI
		--	
		-------------------------------------------------------------------
		createUI = function(modData)
			logFn("files = ", modData.userSelection.files)
			for entry in eatSetContainer(global.misc.getValue(modData.userSelection, {}).files):_pairs() do
				if (#getValue(getAsSet(entry.modelFiles), {}) > 0) then
					entry.defaultIndex = (entry.default == false) and 1 or 0
					modData.myParams.createButton(entry)
				end
			end
		end,
		-------------------------------------------------------------------
		--	run
		--	
		-------------------------------------------------------------------
		run = function(modData)
			local intern = modData.modutilsIntern
			local hideFiles = eatSetEx.new()
			for entry in eatSetContainer(global.misc.getValue(modData.userSelection, {}).files):_pairs() do
				logFn("entry = ", entry)
				logFn("intern.userParams = ", intern.userParams)
				if ((intern.userParams[entry.key] or 0) > 1) then
					hideFiles:append(global.misc.getValue(entry.modelFiles, {}))
				end
			end
			global.mod.hideObject(hideFiles)
		end,
	},
}

local mt = {
	__index = function(self, key)
		for i, class in ipairs({libResult, metaClass}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
}

return setmetatable({}, mt)