--- materialshandler
-- @author Enno Sylvester
-- @copyright 2023
-- @module eatglobal.module.meshhandler

local historie = [[
	Version 1.1:
	 - Release
]]

local versionsData = {
	majorVersion = 1,
	minorVersion = 1,
}

------------------------------------------------------------------------------------
--	Variablen gesetzt in "postLoadFn".
------------------------------------------------------------------------------------
local global
local getVar
local setVar
local logFn
local logFnTraceback
local math
local table

local getValue
------------------------------------------------------------------------------------
--	Vordefinition der Rückgabetabelle.
------------------------------------------------------------------------------------
local libResult

------------------------------------------------------------------------------------
--	Unsere Rückgabetabelle.
------------------------------------------------------------------------------------
libResult = {
	__debug__ = false,
	historie = historie,
	------------------------------------------------------------------------------------
	--	postLoadFn - Automatischer Aufruf durch eatglobal nach Laden der Bibliothek.
	------------------------------------------------------------------------------------
	postLoadFn = function(params)
		-------------------------------------------------------------------
		--	Lokale Variable "global", etc. zuweisen.
		-------------------------------------------------------------------
		global = params.eatglobal
		getVar = params.getVar
		setVar = params.setVar
		logFn = params.logFn
		logFnTraceback = params.logFnTraceback
		libResult.version = tostring(eatVersion.new(versionsData))
		math = global.math
		table = global.table
		
		getValue = global.misc.getValue
	end,
}

------------------------------------------------------------------------------------
--	meta-class
------------------------------------------------------------------------------------
local metaClass
metaClass = {
	metaData = {
		name = "meshhandler",
		library = "modutils",
		priority = 8,
	},
	------------------------------------------------------------------------------------
	--	init
	--	
	------------------------------------------------------------------------------------
	init = function()
	end,
	------------------------------------------------------------------------------------
	--	table handler
	------------------------------------------------------------------------------------
	handler = {
		------------------------------------------------------------------------------------
		--	createUI
		--	
		------------------------------------------------------------------------------------
		createUI = function(params)
			local handleMeshData = table.openKey(params, "userSelection.handleMeshData", false, true)
			if (type(handleMeshData) == "table") then
				local meshData = table.openKey(params, "modutilsIntern.moduleData.meshhandler.meshData", true, false)
				--logFn("handleMeshData = ", handleMeshData)
				for entry in handleMeshData:_pairs() do
					------------------------------------------------------------------------------------
					--	create internal data
					------------------------------------------------------------------------------------
					local handlingData = {}
					for mdl, mdlData in pairs(entry.data) do
						local mdlHandlingData = {
							lodData = {},
							materialData = mdlData.materials,
						}
						
						for lod in eatSetContainer(mdlData.lods):_pairs() do
							local lodData = {}
							for _meshData in eatSetContainer(lod):_pairs() do
								for meshId in eatSetContainer(_meshData.meshIds):_pairs() do
									lodData[meshId] = _meshData.data
								end
							end
							table.insert(mdlHandlingData.lodData, lodData)
						end
						handlingData[mdl] = mdlHandlingData
					end
					meshData[entry.paramsEntry.key] = handlingData
					logFn("handlingData = ", handlingData)
					------------------------------------------------------------------------------------
					--	create UI
					------------------------------------------------------------------------------------
					if (entry.paramsEntry.uiType == "BUTTON") then
						params.myParams.createButton(entry.paramsEntry)
					elseif (entry.paramsEntry.uiType == "SLIDER") then
						params.myParams.createSlider(entry.paramsEntry)
					else
						params.myParams.createCombobox(entry.paramsEntry)
					end
				end
			end
		end,
		------------------------------------------------------------------------------------
		--	run
		--	
		------------------------------------------------------------------------------------
		run = function(params)
			------------------------------------------------------------------------------------
			--	Tool(s)
			------------------------------------------------------------------------------------
			--	doMeshes
			--	
			------------------------------------------------------------------------------------
			local function doMeshes(data, paramsSelection, meshData)
				------------------------------------------------------------------------------------
				--	handleMaterials
				------------------------------------------------------------------------------------
				local function handleMaterials(params)
					if ((type(params.materialData) == "table") and (type(params.root.materials) == "table")) then
						local selectedData = params.materialData[math.clamp(params.paramsSelection - 1, 1, #params.materialData)]
						for mData in eatSetContainer(selectedData):_pairs() do
							for i, meshMaterial in ipairs(params.root.materials) do
								if (meshMaterial == mData[1]) then
									params.root.materials[i] = mData[2]
								end
							end
						end
					end
				end
				------------------------------------------------------------------------------------
				--	handleGroup
				------------------------------------------------------------------------------------
				local function handleGroup(params)
					params.currentId = params.currentId + 1
					local myId = params.currentId
					--	handle materials
					handleMaterials(params)
					--	handle meshes
					if (type(params.lodData) == "table") then
						local lodData = params.lodData[myId]
						if (type(lodData) == "table") then
							local selectedData = lodData[math.clamp(params.paramsSelection - 1, 1, #lodData)]
							for key, value in pairs(selectedData) do
								params.root[key] = value
							end
						end
					end
					
					
					if (type(params.root.children) == "table") then
						for i, child in ipairs(params.root.children) do
							params.currentId = handleGroup({
									root = child,
									currentId = params.currentId,
									parentId = myId,
									lodData = params.lodData,
									materialData = params.materialData,
									paramsSelection = params.paramsSelection,
								} )
						end
					end
					return params.currentId
				end
				
				------------------------------------------------------------------------------------
				------------------------------------------------------------------------------------
				logFn("meshData = ", meshData)
				if (type(data.lods) == "table") then
					for i, lod in ipairs(data.lods) do
						if (type(lod.node) == "table") then
							local lodData = meshData.lodData[i]
							if ((type(meshData.materialData) == "table") or (type(lodData) == "table")) then
								handleGroup({
										root = lod.node,
										currentId = -1,
										parentId = -1,
										lodData = lodData,
										materialData = meshData.materialData,
										paramsSelection = paramsSelection,
									} )
							end
						end
					end
				end
			end
			------------------------------------------------------------------------------------
			--	Hier geht's los. :)
			------------------------------------------------------------------------------------
			local handleMeshData = table.openKey(params, "userSelection.handleMeshData", false, true)
			if (type(handleMeshData) == "table") then
				local meshData = table.openKey(params, "modutilsIntern.moduleData.meshhandler.meshData", false, false)
				if (type(meshData) == "table") then
					local userParams = params.modutilsIntern.userParams
					for key, handlingData in pairs(meshData) do
						local uSelection = userParams[key] or 1
						if (uSelection > 1) then
							global.mod.addModifier("loadModel",
								function(fileName, data)
									for _meshFile, _meshData in pairs(handlingData) do
										if string.ends(fileName, _meshFile) then
											doMeshes(data, userParams[key], _meshData)
											logFn("data.lods = ", data.lods)
											break
										end
									end
									return data
								end,
								metaClass.metaData.priority
							)
						end
					end
				end
			end
		end,
	},
}

return setmetatable({}, {
	__index = function(self, key)
		for i, class in ipairs({libResult, metaClass}) do
			local value = class[key]
			if (value ~= nil) then
				return value
			end
		end
	end,
	__newindex = function(self, key, value) end,
	__metatable = false,
})
